# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import random
import socket

import asynctest
import docker
import mock
import requests
from pytest import mark
from pytest import raises

from paasta_tools import mesos
from paasta_tools import mesos_tools
from paasta_tools import utils
from paasta_tools.utils import PaastaColors


def test_filter_running_tasks():
    tasks = [
        {"id": 1, "state": "TASK_RUNNING", "framework": {"active": True}},
        {"id": 2, "state": "TASK_FAILED", "framework": {"active": True}},
    ]
    running = mesos_tools.filter_running_tasks(tasks)
    assert len(running) == 1
    assert running[0]["id"] == 1


def test_filter_not_running_tasks():
    tasks = [{"id": 1, "state": "TASK_RUNNING"}, {"id": 2, "state": "TASK_FAILED"}]
    not_running = mesos_tools.filter_not_running_tasks(tasks)
    assert len(not_running) == 1
    assert not_running[0]["id"] == 2


@mark.parametrize(
    "test_case",
    [[0, 0], [10, 1 + 10]],  # 1 running task, 10 non-running taks (truncated)
)
def test_status_mesos_tasks_verbose(test_case):
    tail_lines, expected_format_tail_call_count = test_case
    filter_string = "fake--service.fake--instance"

    with asynctest.patch(
        "paasta_tools.mesos_tools.get_cached_list_of_running_tasks_from_frameworks",
        autospec=True,
        return_value=[{"id": filter_string}],
    ), asynctest.patch(
        "paasta_tools.mesos_tools.get_cached_list_of_not_running_tasks_from_frameworks",
        autospec=True,
    ) as get_cached_list_of_not_running_tasks_from_frameworks_patch, asynctest.patch(
        "paasta_tools.mesos_tools.format_running_mesos_task_row", autospec=True
    ) as format_running_mesos_task_row_patch, asynctest.patch(
        "paasta_tools.mesos_tools.format_non_running_mesos_task_row", autospec=True
    ) as format_non_running_mesos_task_row_patch, asynctest.patch(
        "paasta_tools.mesos_tools.format_stdstreams_tail_for_task", autospec=True
    ) as format_stdstreams_tail_for_task_patch:

        template_task_return = {
            "id": filter_string,
            "statuses": [{"timestamp": "##########"}],
            "state": "NOT_RUNNING",
        }
        non_running_mesos_tasks = []
        for _ in range(
            15
        ):  # exercise the code that sorts/truncates the list of non running tasks
            task_return = template_task_return.copy()
            task_return["statuses"][0]["timestamp"] = str(
                1457109986 + random.randrange(-60 * 60 * 24, 60 * 60 * 24)
            )
            non_running_mesos_tasks.append(task_return)
        get_cached_list_of_not_running_tasks_from_frameworks_patch.return_value = (
            non_running_mesos_tasks
        )

        format_running_mesos_task_row_patch.return_value = [
            "id",
            "host",
            "mem",
            "cpu",
            "time",
        ]
        format_non_running_mesos_task_row_patch.return_value = [
            "id",
            "host",
            "time",
            "state",
        ]
        format_stdstreams_tail_for_task_patch.return_value = ["tail"]

        actual = mesos_tools.status_mesos_tasks_verbose(
            filter_string=filter_string,
            get_short_task_id=mock.sentinel.get_short_task_id,
            tail_lines=tail_lines,
        )
        assert "Running Tasks" in actual
        assert "Non-Running Tasks" in actual
        format_running_mesos_task_row_patch.assert_called_once_with(
            {"id": filter_string}, mock.sentinel.get_short_task_id
        )
        assert (
            format_non_running_mesos_task_row_patch.call_count == 10
        )  # maximum n of tasks we display
        assert (
            format_stdstreams_tail_for_task_patch.call_count
            == expected_format_tail_call_count
        )


@mark.asyncio
async def test_get_cpu_usage_good():
    fake_task = mock.create_autospec(mesos.task.Task)
    fake_task.cpu_limit = asynctest.CoroutineMock(return_value=0.35)
    fake_duration = 100
    fake_task.stats = asynctest.CoroutineMock(
        return_value={"cpus_system_time_secs": 2.5, "cpus_user_time_secs": 0.0}
    )
    current_time = datetime.datetime.now()
    fake_task.__getitem__.return_value = [
        {
            "state": "TASK_RUNNING",
            "timestamp": int(current_time.strftime("%s")) - fake_duration,
        }
    ]
    with asynctest.patch(
        "paasta_tools.mesos_tools.datetime.datetime", autospec=True
    ) as mock_datetime:
        mock_datetime.now.return_value = current_time
        actual = await mesos_tools.get_cpu_usage(fake_task)
    assert "10.0%" == actual


@mark.asyncio
async def test_get_cpu_usage_bad():
    fake_task = mock.create_autospec(mesos.task.Task)
    fake_task.cpu_limit = asynctest.CoroutineMock(return_value=1.1)
    fake_duration = 100
    fake_task.stats = asynctest.CoroutineMock(
        return_value={"cpus_system_time_secs": 50.0, "cpus_user_time_secs": 50.0}
    )
    current_time = datetime.datetime.now()
    fake_task.__getitem__.return_value = [
        {
            "state": "TASK_RUNNING",
            "timestamp": int(current_time.strftime("%s")) - fake_duration,
        }
    ]
    with asynctest.patch(
        "paasta_tools.mesos_tools.datetime.datetime", autospec=True
    ) as mock_datetime:
        mock_datetime.now.return_value = current_time
        actual = await mesos_tools.get_cpu_usage(fake_task)
    assert PaastaColors.red("100.0%") in actual


@mark.asyncio
async def test_get_cpu_usage_handles_missing_stats():
    fake_task = mock.create_autospec(mesos.task.Task)
    fake_task.cpu_limit = asynctest.CoroutineMock(return_value=1.1)
    fake_duration = 100
    fake_task.stats = asynctest.CoroutineMock(return_value={})
    fake_task.__getitem__.return_value = [
        {
            "state": "TASK_RUNNING",
            "timestamp": int(datetime.datetime.now().strftime("%s")) - fake_duration,
        }
    ]
    actual = await mesos_tools.get_cpu_usage(fake_task)
    assert "0.0%" in actual


@mark.asyncio
async def test_get_mem_usage_good():
    fake_task = mock.create_autospec(mesos.task.Task)
    fake_task.rss = asynctest.CoroutineMock(return_value=1024 * 1024 * 10)
    fake_task.mem_limit = asynctest.CoroutineMock(return_value=1024 * 1024 * 10 * 10)
    actual = await mesos_tools.get_mem_usage(fake_task)
    assert actual == "10/100MB"


@mark.asyncio
async def test_get_mem_usage_bad():
    fake_task = mock.create_autospec(mesos.task.Task)
    fake_task.rss = asynctest.CoroutineMock(return_value=1024 * 1024 * 100)
    fake_task.mem_limit = fake_task.rss
    actual = await mesos_tools.get_mem_usage(fake_task)
    assert actual == PaastaColors.red("100/100MB")


@mark.asyncio
async def test_get_mem_usage_divide_by_zero():
    fake_task = mock.create_autospec(mesos.task.Task)
    fake_task.rss = asynctest.CoroutineMock(return_value=1024 * 1024 * 10)
    fake_task.mem_limit = asynctest.CoroutineMock(return_value=0)
    actual = await mesos_tools.get_mem_usage(fake_task)
    assert actual == "Undef"


def test_get_zookeeper_config():
    zk_hosts = "1.1.1.1:1111,2.2.2.2:2222,3.3.3.3:3333"
    zk_path = "fake_path"
    fake_state = {"flags": {"zk": f"zk://{zk_hosts}/{zk_path}"}}
    expected = {"hosts": zk_hosts, "path": zk_path}
    assert mesos_tools.get_zookeeper_config(fake_state) == expected


def test_get_mesos_leader():
    fake_url = "http://93.184.216.34:5050"
    with asynctest.patch(
        "paasta_tools.mesos_tools.get_mesos_master", autospec=True
    ) as mock_get_master, asynctest.patch(
        "paasta_tools.mesos_tools.socket.gethostbyaddr", autospec=True
    ) as mock_gethostbyaddr, asynctest.patch(
        "paasta_tools.mesos_tools.socket.getfqdn", autospec=True
    ) as mock_getfqdn:
        mock_master = mock.Mock()
        mock_master.host = fake_url
        mock_get_master.return_value = mock_master
        mock_gethostbyaddr.return_value = "example.org"
        mock_getfqdn.return_value = "example.org"
        assert mesos_tools.get_mesos_leader() == "example.org"


def test_get_mesos_leader_socket_error():
    fake_url = "http://93.184.216.34:5050"
    with asynctest.patch(
        "paasta_tools.mesos_tools.get_mesos_master", autospec=True
    ) as mock_get_master, asynctest.patch(
        "paasta_tools.mesos_tools.socket.gethostbyaddr",
        side_effect=socket.error,
        autospec=True,
    ):
        mock_master = mock.Mock()
        mock_master.host = fake_url
        mock_get_master.return_value = mock_master
        with raises(socket.error):
            mesos_tools.get_mesos_leader()


def test_get_mesos_leader_no_hostname():
    fake_url = "localhost:5050"
    with asynctest.patch(
        "paasta_tools.mesos_tools.get_mesos_master", autospec=True
    ) as mock_get_master:
        mock_master = mock.Mock()
        mock_master.host = fake_url
        mock_get_master.return_value = mock_master
        with raises(ValueError):
            mesos_tools.get_mesos_leader()


@mock.patch(
    "paasta_tools.mesos_tools.get_mesos_config",
    autospec=True,
    return_value={"scheme": "http", "master": "test"},
)
def test_get_mesos_leader_cli_mesosmasterconnectionerror(mock_get_mesos_config):
    with asynctest.patch(
        "paasta_tools.mesos.master.MesosMaster.resolve",
        side_effect=mesos.exceptions.MasterNotAvailableException,
        autospec=True,
    ):
        with raises(mesos.exceptions.MasterNotAvailableException):
            mesos_tools.get_mesos_leader()


@mock.patch("paasta_tools.mesos_tools.KazooClient", autospec=True)
def test_get_number_of_mesos_masters(
    mock_kazoo,
):
    host = "1.1.1.1"
    path = "fake_path"

    zk = mock_kazoo.return_value
    zk.get_children.return_value = ["log_11", "state", "json.info_1", "info_2"]
    assert mesos_tools.get_number_of_mesos_masters(host, path) == 2


@mock.patch("requests.get", autospec=True)
@mock.patch("socket.getfqdn", autospec=True)
def test_get_local_slave_state_connection_error(mock_getfqdn, mock_requests_get):
    fake_request = requests.Request("GET", url="doesnt_matter")
    mock_getfqdn.return_value = "fake_hostname"
    mock_requests_get.side_effect = requests.ConnectionError(
        "fake_message", request=fake_request
    )

    with raises(mesos_tools.MesosSlaveConnectionError):
        mesos_tools.get_local_slave_state()


def test_get_mesos_slaves_grouped_by_attribute():
    fake_value_1 = "fake_value_1"
    fake_value_2 = "fake_value_2"
    fake_slaves = [
        {"hostname": "fake_host_1", "attributes": {"fake_attribute": fake_value_1}},
        {"hostname": "fake_host_2", "attributes": {"fake_attribute": fake_value_2}},
        {"hostname": "fake_host_3", "attributes": {"fake_attribute": fake_value_1}},
        {
            "hostname": "fake_host_4",
            "attributes": {"fake_attribute": "fake_other_value"},
        },
    ]
    expected = {
        "fake_value_1": [
            {"hostname": "fake_host_1", "attributes": {"fake_attribute": fake_value_1}},
            {"hostname": "fake_host_3", "attributes": {"fake_attribute": fake_value_1}},
        ],
        "fake_value_2": [
            {"hostname": "fake_host_2", "attributes": {"fake_attribute": fake_value_2}}
        ],
        "fake_other_value": [
            {
                "hostname": "fake_host_4",
                "attributes": {"fake_attribute": "fake_other_value"},
            }
        ],
    }
    actual = mesos_tools.get_mesos_slaves_grouped_by_attribute(
        fake_slaves, "fake_attribute"
    )
    assert actual == expected


@mock.patch("paasta_tools.mesos_tools.host_passes_blacklist", autospec=True)
def test_filter_mesos_slaves_by_blacklist_when_unfiltered(mock_slave_passes_blacklist):
    mock_slave_passes_blacklist.return_value = True
    slaves = [
        {"hostname": "fake_host_1", "attributes": {"fake_attribute": "fake_value_1"}},
        {"hostname": "fake_host_2", "attributes": {"fake_attribute": "fake_value_1"}},
    ]
    blacklist = []
    whitelist = None
    actual = mesos_tools.filter_mesos_slaves_by_blacklist(
        slaves=slaves, blacklist=blacklist, whitelist=whitelist
    )
    assert mock_slave_passes_blacklist.call_count == 2
    assert actual == slaves


@mock.patch("paasta_tools.mesos_tools.host_passes_blacklist", autospec=True)
def test_filter_mesos_slaves_by_blacklist_when_filtered(mock_slave_passes_blacklist):
    mock_slave_passes_blacklist.return_value = False
    slaves = [
        {"hostname": "fake_host_1", "attributes": {"fake_attribute": "fake_value_1"}},
        {"hostname": "fake_host_2", "attributes": {"fake_attribute": "fake_value_1"}},
    ]
    blacklist = []
    whitelist = None
    actual = mesos_tools.filter_mesos_slaves_by_blacklist(
        slaves=slaves, blacklist=blacklist, whitelist=whitelist
    )
    assert mock_slave_passes_blacklist.call_count == 2
    assert actual == []


def test_get_paasta_execute_docker_healthcheck():
    mock_docker_client = mock.MagicMock(spec_set=docker.APIClient)
    fake_container_id = "fake_container_id"
    fake_mesos_id = "fake_mesos_id"
    fake_container_info = [
        {"Config": {"Env": None}},
        {
            "Config": {
                "Env": ["fake_key1=fake_value1", "MESOS_TASK_ID=fake_other_mesos_id"]
            },
            "Id": "11111",
        },
        {
            "Config": {
                "Env": ["fake_key2=fake_value2", "MESOS_TASK_ID=%s" % fake_mesos_id]
            },
            "Id": fake_container_id,
        },
    ]
    mock_docker_client.containers = mock.MagicMock(
        spec_set=docker.APIClient,
        return_value=["fake_container_1", "fake_container_2", "fake_container_3"],
    )
    mock_docker_client.inspect_container = mock.MagicMock(
        spec_set=docker.APIClient, side_effect=fake_container_info
    )
    assert (
        mesos_tools.get_container_id_for_mesos_id(mock_docker_client, fake_mesos_id)
        == fake_container_id
    )


def test_get_paasta_execute_docker_healthcheck_when_not_found():
    mock_docker_client = mock.MagicMock(spec_set=docker.APIClient)
    fake_mesos_id = "fake_mesos_id"
    fake_container_info = [
        {
            "Config": {
                "Env": ["fake_key1=fake_value1", "MESOS_TASK_ID=fake_other_mesos_id"]
            },
            "Id": "11111",
        },
        {
            "Config": {
                "Env": ["fake_key2=fake_value2", "MESOS_TASK_ID=fake_other_mesos_id2"]
            },
            "Id": "2222",
        },
    ]
    mock_docker_client.containers = mock.MagicMock(
        spec_set=docker.APIClient, return_value=["fake_container_1", "fake_container_2"]
    )
    mock_docker_client.inspect_container = mock.MagicMock(
        spec_set=docker.APIClient, side_effect=fake_container_info
    )
    assert (
        mesos_tools.get_container_id_for_mesos_id(mock_docker_client, fake_mesos_id)
        is None
    )


@mark.parametrize(
    "test_case",
    [
        [
            ["taska", "taskb"],  # test_case0 - OK
            [["outlna1", "outlna2", "errlna1"], ["outlnb1", "errlnb1", "errlnb2"]],
            [
                "taska",
                "outlna1",
                "outlna2",
                "errlna1",
                "taskb",
                "outlnb1",
                "errlnb1",
                "errlnb2",
            ],
            False,
        ],
        [["a"], [1, 2], None, True],  # test_case1 - can't zip different length lists
    ],
)
def test_zip_tasks_verbose_output(test_case):
    table, stdstreams, expected, should_raise = test_case
    result = None
    raised = False
    try:
        result = mesos_tools.zip_tasks_verbose_output(table, stdstreams)
    except ValueError:
        raised = True

    assert raised == should_raise
    assert result == expected


@mark.asyncio
@mark.parametrize(
    "test_case",
    [
        # task_id, file1, file2, nlines, raise_what
        [
            "a_task",  # test_case0 - OK
            ["stdout", [str(x) for x in range(20)]],
            ["stderr", [str(x) for x in range(30)]],
            10,
            None,
        ],
        [
            "a_task",  # test_case1 - OK, short stdout, swapped stdout/stderr
            ["stderr", [str(x) for x in range(30)]],
            ["stdout", ["1", "2"]],
            10,
            None,
        ],
        ["a_task", None, None, 10, mesos.exceptions.MasterNotAvailableException],
        ["a_task", None, None, 10, mesos.exceptions.SlaveDoesNotExist],
        ["a_task", None, None, 10, mesos.exceptions.TaskNotFoundException],
        ["a_task", None, None, 10, mesos.exceptions.FileNotFoundForTaskException],
        ["a_task", None, None, 10, utils.TimeoutError],
    ],
)
async def test_format_stdstreams_tail_for_task(
    test_case,
):
    def gen_mesos_cli_fobj(file_path, file_lines):
        """mesos.cli.cluster.files (0.1.5),
        returns a list of mesos.cli.mesos_file.File
        `File` is an iterator-like object.
        """

        async def _readlines_reverse():
            for line in reversed(file_lines):
                yield line

        fobj = mock.create_autospec(mesos.mesos_file.File)
        fobj.path = file_path
        fobj._readlines_reverse = _readlines_reverse
        return fobj

    def get_short_task_id(task_id):
        return task_id

    def gen_mock_cluster_files(file1, file2, raise_what):
        async def mock_cluster_files(*args, **kwargs):
            # If we're asked to raise a particular exception we do so.
            # .message is set to the exception class name.
            if raise_what:
                raise raise_what(raise_what)
            yield gen_mesos_cli_fobj(file1[0], file1[1])
            yield gen_mesos_cli_fobj(file2[0], file2[1])

        return mock_cluster_files

    def gen_output(task_id, file1, file2, nlines, raise_what):
        error_message = "    " + PaastaColors.red(
            "  couldn't read stdout/stderr for %s (%s)"
        )
        output = []
        if not raise_what:
            files = [file1, file2]
            # reverse sort because stdout is supposed to always come before stderr in the output
            files.sort(key=lambda f: f[0], reverse=True)
            for f in files:
                output.append(
                    "    " + PaastaColors.blue("{} tail for {}".format(f[0], task_id))
                )
                output.extend(f"      {line}" for line in f[1][-nlines:])
        else:
            output.append(error_message % (task_id, raise_what.__name__))
        return output

    task_id, file1, file2, nlines, raise_what = test_case

    mock_cluster_files = gen_mock_cluster_files(file1, file2, raise_what)
    fake_task = {"id": task_id}
    expected = gen_output(task_id, file1, file2, nlines, raise_what)
    with asynctest.patch("paasta_tools.mesos_tools.get_mesos_config", autospec=True):
        with asynctest.patch(
            "paasta_tools.mesos_tools.cluster.get_files_for_tasks",
            mock_cluster_files,
            autospec=None,
        ):
            result = await mesos_tools.format_stdstreams_tail_for_task(
                fake_task, get_short_task_id
            )
            assert result == expected


def test_slave_pid_to_ip():
    ret = mesos_tools.slave_pid_to_ip("slave(1)@10.40.31.172:5051")
    assert ret == "10.40.31.172"


@mark.asyncio
async def test_get_mesos_task_count_by_slave():
    with asynctest.patch(
        "paasta_tools.mesos_tools.get_all_running_tasks", autospec=True
    ) as mock_get_all_running_tasks:
        mock_tron = mock.Mock()
        mock_tron.name = "tron"
        mock_somethingelse = mock.Mock()
        mock_somethingelse.name = "somethingelse"
        mock_task1 = mock.Mock()
        mock_task1.slave = asynctest.CoroutineMock(return_value={"id": "slave1"})
        mock_task1.framework = asynctest.CoroutineMock(return_value=mock_tron)
        mock_task2 = mock.Mock()
        mock_task2.slave = asynctest.CoroutineMock(return_value={"id": "slave1"})
        mock_task2.framework = asynctest.CoroutineMock(return_value=mock_somethingelse)
        mock_task3 = mock.Mock()
        mock_task3.slave = asynctest.CoroutineMock(return_value={"id": "slave1"})
        mock_task3.framework = asynctest.CoroutineMock(return_value=mock_somethingelse)
        mock_task4 = mock.Mock()
        mock_task4.slave = asynctest.CoroutineMock(return_value={"id": "slave2"})
        mock_task4.framework = asynctest.CoroutineMock(return_value=mock_somethingelse)
        mock_tasks = [mock_task1, mock_task2, mock_task3, mock_task4]
        mock_get_all_running_tasks.return_value = mock_tasks
        mock_slave_1 = {
            "id": "slave1",
            "attributes": {"pool": "default"},
            "hostname": "host1",
        }
        mock_slave_2 = {
            "id": "slave2",
            "attributes": {"pool": "default"},
            "hostname": "host2",
        }
        mock_slave_3 = {
            "id": "slave3",
            "attributes": {"pool": "another"},
            "hostname": "host3",
        }
        mock_mesos_state = {"slaves": [mock_slave_1, mock_slave_2, mock_slave_3]}
        ret = await mesos_tools.get_mesos_task_count_by_slave(
            mock_mesos_state, pool="default"
        )
        assert mock_get_all_running_tasks.called
        expected = [
            {"task_counts": mesos_tools.SlaveTaskCount(count=3, slave=mock_slave_1)},
            {"task_counts": mesos_tools.SlaveTaskCount(count=1, slave=mock_slave_2)},
        ]
        assert len(ret) == len(expected) and utils.sort_dicts(ret) == utils.sort_dicts(
            expected
        )
        ret = await mesos_tools.get_mesos_task_count_by_slave(
            mock_mesos_state, pool=None
        )
        assert mock_get_all_running_tasks.called
        expected = [
            {"task_counts": mesos_tools.SlaveTaskCount(count=3, slave=mock_slave_1)},
            {"task_counts": mesos_tools.SlaveTaskCount(count=1, slave=mock_slave_2)},
            {"task_counts": mesos_tools.SlaveTaskCount(count=0, slave=mock_slave_3)},
        ]
        assert len(ret) == len(expected) and utils.sort_dicts(ret) == utils.sort_dicts(
            expected
        )

        # test slaves_list override
        mock_task2 = mock.Mock()
        mock_task2.slave = asynctest.CoroutineMock(return_value={"id": "slave2"})
        mock_task2.framework = asynctest.CoroutineMock(return_value=mock_somethingelse)
        mock_task3 = mock.Mock()
        mock_task3.slave = asynctest.CoroutineMock(return_value={"id": "slave2"})
        mock_task3.framework = asynctest.CoroutineMock(return_value=mock_somethingelse)
        mock_tasks = [mock_task1, mock_task2, mock_task3, mock_task4]
        mock_get_all_running_tasks.return_value = mock_tasks
        mock_slaves_list = [
            {"task_counts": mesos_tools.SlaveTaskCount(count=0, slave=mock_slave_1)},
            {"task_counts": mesos_tools.SlaveTaskCount(count=0, slave=mock_slave_2)},
            {"task_counts": mesos_tools.SlaveTaskCount(count=0, slave=mock_slave_3)},
        ]
        ret = await mesos_tools.get_mesos_task_count_by_slave(
            mock_mesos_state, slaves_list=mock_slaves_list
        )
        expected = [
            {"task_counts": mesos_tools.SlaveTaskCount(count=1, slave=mock_slave_1)},
            {"task_counts": mesos_tools.SlaveTaskCount(count=3, slave=mock_slave_2)},
            {"task_counts": mesos_tools.SlaveTaskCount(count=0, slave=mock_slave_3)},
        ]
        assert len(ret) == len(expected) and utils.sort_dicts(ret) == utils.sort_dicts(
            expected
        )

        # test SlaveDoesNotExist exception handling
        mock_task2.__getitem__ = mock.Mock(side_effect="fakeid")
        mock_task2.slave = asynctest.CoroutineMock(
            return_value=mock.Mock(
                __getitem__=mock.Mock(side_effect=mesos.exceptions.SlaveDoesNotExist)
            )
        )
        # we expect to handle this SlaveDoesNotExist exception gracefully, and continue on to handle other tasks
        mock_tasks = [mock_task1, mock_task2, mock_task3, mock_task4]
        mock_get_all_running_tasks.return_value = mock_tasks
        mock_slaves_list = [
            {"task_counts": mesos_tools.SlaveTaskCount(count=0, slave=mock_slave_1)},
            {"task_counts": mesos_tools.SlaveTaskCount(count=0, slave=mock_slave_2)},
            {"task_counts": mesos_tools.SlaveTaskCount(count=0, slave=mock_slave_3)},
        ]
        ret = await mesos_tools.get_mesos_task_count_by_slave(
            mock_mesos_state, slaves_list=mock_slaves_list
        )
        # we expect mock_slave_2 to only count 2 tasks, as one of them returned a SlaveDoesNotExist exception
        expected = [
            {"task_counts": mesos_tools.SlaveTaskCount(count=1, slave=mock_slave_1)},
            {"task_counts": mesos_tools.SlaveTaskCount(count=2, slave=mock_slave_2)},
            {"task_counts": mesos_tools.SlaveTaskCount(count=0, slave=mock_slave_3)},
        ]
        assert len(ret) == len(expected) and utils.sort_dicts(ret) == utils.sort_dicts(
            expected
        )


def test_get_count_running_tasks_on_slave():
    with asynctest.patch(
        "paasta_tools.mesos_tools.get_mesos_master", autospec=True
    ) as mock_get_master, asynctest.patch(
        "paasta_tools.mesos_tools.get_mesos_task_count_by_slave", autospec=True
    ) as mock_get_mesos_task_count_by_slave:
        mock_master = mock.Mock()
        mock_mesos_state = mock.Mock()
        mock_master.state_summary = asynctest.CoroutineMock(
            func=asynctest.CoroutineMock(),  # https://github.com/notion/a_sync/pull/40
            return_value=mock_mesos_state,
        )
        mock_get_master.return_value = mock_master

        mock_slave_counts = [
            {"task_counts": mock.Mock(count=3, slave={"hostname": "host1"})},
            {"task_counts": mock.Mock(count=0, slave={"hostname": "host2"})},
        ]
        mock_get_mesos_task_count_by_slave.return_value = mock_slave_counts

        assert mesos_tools.get_count_running_tasks_on_slave("host1") == 3
        assert mesos_tools.get_count_running_tasks_on_slave("host2") == 0
        assert mesos_tools.get_count_running_tasks_on_slave("host3") == 0
        assert mock_master.state_summary.called
        mock_get_mesos_task_count_by_slave.assert_called_with(mock_mesos_state)


def _ids(list_of_mocks):
    return {id(mck) for mck in list_of_mocks}


@mark.asyncio
async def test_get_tasks_from_app_id():
    with asynctest.patch(
        "paasta_tools.mesos_tools.get_running_tasks_from_frameworks", autospec=True
    ) as mock_get_running_tasks_from_frameworks:
        mock_task_1 = mock.Mock(
            slave=asynctest.CoroutineMock(return_value={"hostname": "host1"})
        )
        mock_task_2 = mock.Mock(
            slave=asynctest.CoroutineMock(return_value={"hostname": "host2"})
        )
        mock_task_3 = mock.Mock(
            slave=asynctest.CoroutineMock(return_value={"hostname": "host2.domain"})
        )
        mock_get_running_tasks_from_frameworks.return_value = [
            mock_task_1,
            mock_task_2,
            mock_task_3,
        ]

        ret = await mesos_tools.get_tasks_from_app_id("app_id")
        mock_get_running_tasks_from_frameworks.assert_called_with("app_id")
        expected = [mock_task_1, mock_task_2, mock_task_3]
        assert len(expected) == len(ret) and _ids(ret) == _ids(expected)

        ret = await mesos_tools.get_tasks_from_app_id("app_id", slave_hostname="host2")
        mock_get_running_tasks_from_frameworks.assert_called_with("app_id")
        expected = [mock_task_2, mock_task_3]
        assert len(expected) == len(ret) and _ids(ret) == _ids(expected)


@mark.asyncio
async def test_get_task():
    with asynctest.patch(
        "paasta_tools.mesos_tools.get_running_tasks_from_frameworks", autospec=True
    ) as mock_get_running_tasks_from_frameworks:
        mock_task_1 = {"id": "123"}
        mock_task_2 = {"id": "789"}
        mock_task_3 = {"id": "789"}
        mock_get_running_tasks_from_frameworks.return_value = [
            mock_task_1,
            mock_task_2,
            mock_task_3,
        ]
        ret = await mesos_tools.get_task("123", app_id="app_id")
        mock_get_running_tasks_from_frameworks.assert_called_with("app_id")
        assert ret == mock_task_1

        with raises(mesos_tools.TaskNotFound):
            await mesos_tools.get_task("111", app_id="app_id")

        with raises(mesos_tools.TooManyTasks):
            await mesos_tools.get_task("789", app_id="app_id")


@mark.asyncio
async def test_filter_task_by_hostname():
    mock_task = mock.Mock(
        slave=asynctest.CoroutineMock(return_value={"hostname": "host1"})
    )
    assert await mesos_tools.filter_task_by_hostname(mock_task, "host1")
    assert not await mesos_tools.filter_task_by_hostname(mock_task, "host2")


def test_filter_task_by_task_id():
    mock_task = {"id": "123"}
    assert mesos_tools.filter_task_by_task_id(mock_task, "123")
    assert not mesos_tools.filter_task_by_task_id(mock_task, "456")


def test_get_all_tasks_from_state():
    mock_task_1 = mock.Mock()
    mock_task_2 = mock.Mock()
    mock_task_3 = mock.Mock()
    mock_task_4 = mock.Mock()
    mock_state = {
        "frameworks": [{"tasks": [mock_task_1, mock_task_2]}, {"tasks": [mock_task_3]}],
        "orphan_tasks": [mock_task_4],
    }
    ret = mesos_tools.get_all_tasks_from_state(mock_state)
    expected = [mock_task_1, mock_task_2, mock_task_3]
    assert len(ret) == len(expected) and ret == expected

    ret = mesos_tools.get_all_tasks_from_state(mock_state, include_orphans=True)
    expected = [mock_task_1, mock_task_2, mock_task_3, mock_task_4]
    assert len(ret) == len(expected) and ret == expected


@mark.asyncio
async def test_get_running_tasks_from_frameworks():
    with asynctest.patch(
        "paasta_tools.mesos_tools.get_current_tasks", autospec=True
    ) as mock_get_current_tasks, asynctest.patch(
        "paasta_tools.mesos_tools.filter_running_tasks", autospec=True
    ) as mock_filter_running_tasks:
        ret = await mesos_tools.get_running_tasks_from_frameworks(job_id="")
        mock_get_current_tasks.assert_called_with("")
        mock_filter_running_tasks.assert_called_with(
            mock_get_current_tasks.return_value
        )
        assert ret == mock_filter_running_tasks.return_value


@mark.asyncio
async def test_get_all_running_tasks():
    with asynctest.patch(
        "paasta_tools.mesos_tools.get_current_tasks", autospec=True
    ) as mock_get_current_tasks, asynctest.patch(
        "paasta_tools.mesos_tools.filter_running_tasks", autospec=True
    ) as mock_filter_running_tasks, asynctest.patch(
        "paasta_tools.mesos_tools.get_mesos_master", autospec=True
    ) as mock_get_mesos_master:
        mock_task_1 = mock.Mock()
        mock_task_2 = mock.Mock()
        mock_task_3 = mock.Mock()

        mock_get_current_tasks.return_value = [mock_task_1, mock_task_2]
        mock_orphan_tasks = asynctest.CoroutineMock(return_value=[mock_task_3])
        mock_mesos_master = mock.Mock(orphan_tasks=mock_orphan_tasks)
        mock_get_mesos_master.return_value = mock_mesos_master

        ret = await mesos_tools.get_all_running_tasks()
        mock_get_current_tasks.assert_called_with("")
        mock_filter_running_tasks.assert_called_with(
            [mock_task_1, mock_task_2, mock_task_3]
        )
        assert ret == mock_filter_running_tasks.return_value


@mark.asyncio
async def test_get_non_running_tasks_from_frameworks():
    with asynctest.patch(
        "paasta_tools.mesos_tools.get_current_tasks", autospec=True
    ) as mock_get_current_tasks, asynctest.patch(
        "paasta_tools.mesos_tools.filter_not_running_tasks", autospec=True
    ) as mock_filter_not_running_tasks:
        ret = await mesos_tools.get_non_running_tasks_from_frameworks(job_id="")
        mock_get_current_tasks.assert_called_with("")
        mock_filter_not_running_tasks.assert_called_with(
            mock_get_current_tasks.return_value
        )
        assert ret == mock_filter_not_running_tasks.return_value


@mark.asyncio
async def test_get_current_tasks():
    with asynctest.patch(
        "paasta_tools.mesos_tools.get_mesos_master", autospec=True
    ) as mock_get_mesos_master:
        mock_task_1 = mock.Mock()
        mock_task_2 = mock.Mock()
        mock_tasks = asynctest.CoroutineMock(return_value=[mock_task_1, mock_task_2])
        mock_mesos_master = mock.Mock(tasks=mock_tasks)
        mock_get_mesos_master.return_value = mock_mesos_master

        expected = [mock_task_1, mock_task_2]
        ret = await mesos_tools.get_current_tasks("")
        assert ret == expected and len(ret) == len(expected)


def test_mesos_services_running_here():
    with mock.patch(
        "paasta_tools.mesos_tools.get_local_slave_state", autospec=True
    ) as mock_get_local_slave_state:
        mock_state = {
            "frameworks": [
                {
                    "name": "marathon2",
                    "executors": [
                        {
                            "id": "thing.main",
                            "resources": {"ports": "[31062-31062]"},
                            "tasks": [{"state": "TASK_RUNNING"}],
                        },
                        {"id": "thing.another", "tasks": [{"state": "TASK_LOST"}]},
                    ],
                },
                {
                    "name": "tron",
                    "executors": [
                        {
                            "id": "c.main",
                            "resources": {},
                            "tasks": [{"state": "TASK_RUNNING"}],
                        },
                        {
                            "id": "c.another",
                            "resources": {},
                            "tasks": [{"state": "TASK_RUNNING"}],
                        },
                    ],
                },
            ]
        }
        mock_get_local_slave_state.return_value = {}
        assert (
            mesos_tools.mesos_services_running_here(
                lambda _: True, lambda id_str: id_str.split(".")
            )
            == []
        )

        mock_get_local_slave_state.return_value = mock_state
        expected = [
            ("thing", "main", 31062),
            ("c", "main", None),
            ("c", "another", None),
        ]
        assert (
            mesos_tools.mesos_services_running_here(
                lambda _: True, lambda id_str: id_str.split(".")
            )
            == expected
        )

        mock_fw_filter = mock.Mock(side_effect=[True, False])
        expected = [("thing", "main", 31062)]
        assert (
            mesos_tools.mesos_services_running_here(
                mock_fw_filter, lambda id_str: id_str.split(".")
            )
            == expected
        )

        mock_parse_service_instance_from_executor_id = mock.Mock(
            side_effect=[("thing", "main"), ValueError, ("c", "another")]
        )
        expected = [("thing", "main", 31062), ("c", "another", None)]
        assert (
            mesos_tools.mesos_services_running_here(
                lambda _: True, mock_parse_service_instance_from_executor_id
            )
            == expected
        )



import mock
from pytest import raises

from paasta_tools.delete_kubernetes_deployments import get_deployment_names_from_list
from paasta_tools.delete_kubernetes_deployments import main
from paasta_tools.utils import InvalidJobNameError


def test_main():
    with mock.patch(
        "paasta_tools.delete_kubernetes_deployments.get_deployment_names_from_list",
        autospec=True,
    ) as mock_get_deployment_names_from_list, mock.patch(
        "paasta_tools.delete_kubernetes_deployments.delete_deployment",
        autospec=True,
    ) as mock_delete_deployment, mock.patch(
        "paasta_tools.delete_kubernetes_deployments.KubeClient", autospec=True
    ) as mock_kube_client, mock.patch(
        "paasta_tools.delete_kubernetes_deployments.ensure_namespace",
        autospec=True,
    ) as mock_ensure_namespace:
        # Test main() success
        mock_get_deployment_names_from_list.return_value = ["fake_pcm_deployment"]
        with raises(SystemExit) as e:
            main(args=["fake_pcm_deployment"])
        assert e.value.code == 0
        assert mock_ensure_namespace.called
        mock_delete_deployment.assert_called_with(
            kube_client=mock_kube_client.return_value,
            deployment_name="fake_pcm_deployment",
            namespace="paasta",
        )

        # Test main() failed
        mock_delete_deployment.side_effect = Exception("Delete Error")
        with raises(SystemExit) as e:
            main(args=["fake_pcm_deployment"])
        assert e.value.code == 1


def test_get_deployment_names_from_list():
    with mock.patch(
        "paasta_tools.delete_kubernetes_deployments.decompose_job_id", autospec=True
    ) as mock_decompose_job_id:
        # Test get_deployment_names_from_list() success
        mock_decompose_job_id.return_value = (
            "fake-service",
            "fake_instance",
            "fake_hash",
            "fake_hash",
        )
        output = get_deployment_names_from_list(["fake-service.fake_instance"])
        assert output[0] == "fake-service-fake--instance"

        # Test get_deployment_names_from_list() failed
        mock_decompose_job_id.side_effect = InvalidJobNameError()
        with raises(SystemExit) as e:
            get_deployment_names_from_list(["fake-service.fake_instance"])
        assert e.value.code == 1



import mock
import pytest

from paasta_tools import firewall_logging


@mock.patch.object(firewall_logging, "lookup_service_instance_by_ip")
def test_syslog_to_paasta_log(mock_lookup_service_instance_by_ip, mock_log):
    syslog_data = fake_syslog_data("my-hostname", SRC="1.2.3.4")
    mock_lookup_service_instance_by_ip.return_value = ("myservice", "myinstance")

    firewall_logging.syslog_to_paasta_log(syslog_data, "my-cluster")

    assert mock_log.mock_calls == [
        mock.call(
            service="myservice",
            component="security",
            level="debug",
            cluster="my-cluster",
            instance="myinstance",
            line="my-hostname: my-prefix IN=docker0 SRC=1.2.3.4",
        )
    ]


@mock.patch.object(firewall_logging, "lookup_service_instance_by_ip")
def test_syslog_to_paasta_log_no_container(
    mock_lookup_service_instance_by_ip, mock_log
):
    syslog_data = fake_syslog_data("my-hostname", SRC="1.2.3.4")
    mock_lookup_service_instance_by_ip.return_value = (None, None)
    firewall_logging.syslog_to_paasta_log(syslog_data, "my-cluster")
    assert mock_log.mock_calls == []


def test_parse_syslog_undecodable():
    assert (
        firewall_logging.parse_syslog(b"<4>Jun  6 07:52:38 myhost \xba~\xa6r") is None
    )


@pytest.mark.parametrize(
    "syslog_data",
    [
        "<4>Jun  6 07:52:38 myhost someothermessage: hello world",
        "<4>Jun  6 07:52:38 myhost kernel: hello world",
        "<4>Jun  6 07:52:38 myhost kernel [0.0]: hello world",
    ],
)
@mock.patch.object(firewall_logging, "lookup_service_instance_by_ip")
def test_syslog_to_paasta_log_bad_message(
    mock_lookup_service_instance_by_ip, syslog_data, mock_log
):
    firewall_logging.syslog_to_paasta_log(syslog_data.encode(), "my-cluster")
    assert mock_lookup_service_instance_by_ip.mock_calls == []
    assert mock_log.mock_calls == []


@mock.patch.object(
    firewall_logging,
    "services_running_here",
    return_value=[
        ("service1", "instance1", "00:00:00:00:00", "1.1.1.1"),
        ("service1", "instance2", "00:00:00:00:00", "2.2.2.2"),
    ],
)
@mock.patch.object(firewall_logging, "log")
def test_lookup_service_instance_by_ip(my_mock_log, mock_services_running_here):
    assert firewall_logging.lookup_service_instance_by_ip("1.1.1.1") == (
        "service1",
        "instance1",
    )
    assert firewall_logging.lookup_service_instance_by_ip("2.2.2.2") == (
        "service1",
        "instance2",
    )
    assert firewall_logging.lookup_service_instance_by_ip("3.3.3.3") == (None, None)
    assert my_mock_log.info.mock_calls == [
        mock.call("Unable to find container for ip 3.3.3.3")
    ]


def test_parse_args():
    assert firewall_logging.parse_args([]).listen_host == "127.0.0.1"
    assert firewall_logging.parse_args([]).listen_port == 1516
    assert firewall_logging.parse_args([]).verbose is False
    assert firewall_logging.parse_args(["-v"]).verbose is True
    assert firewall_logging.parse_args(["-p", "1234"]).listen_port == 1234
    assert firewall_logging.parse_args(["-l", "0.0.0.0"]).listen_host == "0.0.0.0"


@mock.patch.object(firewall_logging, "logging")
def test_setup_logging(logging_mock):
    firewall_logging.setup_logging(True)
    assert logging_mock.basicConfig.mock_calls == [mock.call(level=logging_mock.DEBUG)]


@mock.patch.object(firewall_logging, "MultiUDPServer")
def test_run_server(udpserver_mock):
    firewall_logging.run_server("myhost", 1234)
    assert udpserver_mock.mock_calls == [
        mock.call(("myhost", 1234), firewall_logging.SyslogUDPHandler),
        mock.call().serve_forever(),
    ]


@mock.patch.object(firewall_logging, "logging")
@mock.patch.object(firewall_logging, "MultiUDPServer")
@mock.patch.object(firewall_logging, "signal")
def test_main_single_worker(signal_mock, udpserver_mock, logging_mock):
    firewall_logging.main(["-w", "1"])
    assert logging_mock.basicConfig.mock_calls == [
        mock.call(level=logging_mock.WARNING)
    ]
    assert udpserver_mock.mock_calls == [
        mock.call(("127.0.0.1", 1516), firewall_logging.SyslogUDPHandler),
        mock.call().serve_forever(),
    ]


@mock.patch.object(firewall_logging, "logging")
@mock.patch.object(firewall_logging, "MultiUDPServer")
@mock.patch.object(firewall_logging.os, "fork", return_value=0)
@mock.patch.object(firewall_logging, "signal")
def test_main_two_workers(signal_mock, fork_mock, udpserver_mock, logging_mock):
    firewall_logging.main(["-w", "2"])
    assert logging_mock.basicConfig.mock_calls == [
        mock.call(level=logging_mock.WARNING)
    ]
    assert udpserver_mock.mock_calls == [
        mock.call(("127.0.0.1", 1516), firewall_logging.SyslogUDPHandler),
        mock.call().serve_forever(),
        mock.call(("127.0.0.1", 1516), firewall_logging.SyslogUDPHandler),
        mock.call().serve_forever(),
    ]


def fake_syslog_data(hostname, **kwargs):
    prefix = (
        f"<4>Jun  6 07:52:38 {hostname} kernel: [2736265.340132] my-prefix IN=docker0 "
    )
    fields_str = " ".join(map("=".join, kwargs.items()))
    return (prefix + fields_str + " \n").encode()


@pytest.fixture
def mock_log():
    with mock.patch.object(firewall_logging, "_log", autospec=True) as mock_log:
        yield mock_log



# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock

import paasta_tools.flink_tools as flink_tools


def test_get_flink_ingress_url_root():
    assert (
        flink_tools.get_flink_ingress_url_root("mycluster", False)
        == "http://flink.k8s.mycluster.paasta:31080/"
    )


@mock.patch("requests.Response", autospec=True)
@mock.patch("paasta_tools.flink_tools.requests.get", autospec=True)
@mock.patch("paasta_tools.flink_tools.get_cr", autospec=True)
def test_curl_flink_endpoint_error(
    mock_get_cr,
    mock_requests_get,
    mock_response,
):
    mock_get_cr.return_value = {
        "metadata": {
            "labels": {"paasta.yelp.com/cluster": "mocked"},
            "annotations": {
                "flink.yelp.com/dashboard_url": "http://flink.k8s.test_cluster.paasta:31080/kurupt-7f5cfd8ffc"
            },
        }
    }
    mock_requests_get.return_value = mock_response
    mock_response.ok = False
    mock_response.status_code = 401
    mock_response.reason = "Unauthorized"
    mock_response.text = "401 Authorization Required"

    service = "kurupt"
    instance = "main"
    response = flink_tools.curl_flink_endpoint(
        flink_tools.cr_id(service, instance), "overview"
    )

    assert response == {
        "status": 401,
        "error": "Unauthorized",
        "text": "401 Authorization Required",
    }


@mock.patch("requests.Response", autospec=True)
@mock.patch("paasta_tools.flink_tools.requests.get", autospec=True)
@mock.patch("paasta_tools.flink_tools.get_cr", autospec=True)
def test_curl_flink_endpoint_overview(
    mock_get_cr,
    mock_requests_get,
    mock_response,
):
    mock_get_cr.return_value = {
        "metadata": {
            "labels": {"paasta.yelp.com/cluster": "mocked"},
            "annotations": {
                "flink.yelp.com/dashboard_url": "http://flink.k8s.test_cluster.paasta:31080/kurupt-7f5cfd8ffc"
            },
        }
    }
    mock_requests_get.return_value = mock_response
    mock_response.json.return_value = {
        "taskmanagers": 5,
        "slots-total": 25,
        "slots-available": 3,
        "jobs-running": 1,
        "jobs-finished": 0,
        "jobs-cancelled": 0,
        "jobs-failed": 0,
    }

    service = "kurupt"
    instance = "main"
    overview = flink_tools.curl_flink_endpoint(
        flink_tools.cr_id(service, instance), "overview"
    )

    assert overview == {
        "taskmanagers": 5,
        "slots-total": 25,
        "slots-available": 3,
        "jobs-running": 1,
        "jobs-finished": 0,
        "jobs-cancelled": 0,
        "jobs-failed": 0,
    }


@mock.patch("requests.Response", autospec=True)
@mock.patch("paasta_tools.flink_tools.requests.get", autospec=True)
@mock.patch("paasta_tools.flink_tools.get_cr", autospec=True)
def test_curl_flink_endpoint_config(
    mock_get_cr,
    mock_requests_get,
    mock_response,
):
    mock_get_cr.return_value = {
        "metadata": {
            "labels": {"paasta.yelp.com/cluster": "mocked"},
            "annotations": {
                "flink.yelp.com/dashboard_url": "http://flink.k8s.test_cluster.paasta:31080/kurupt-7f5cfd8ffc"
            },
        }
    }
    mock_requests_get.return_value = mock_response
    mock_response.json.return_value = {
        "refresh-interval": 3000,
        "timezone-name": "Coordinated Universal Time",
        "timezone-offset": 0,
        "flink-version": "1.13.5",
        "flink-revision": "0ff28a7 @ 2021-12-14T23:26:04+01:00",
        "features": {"web-submit": True},
    }

    service = "kurupt"
    instance = "main"
    config = flink_tools.curl_flink_endpoint(
        flink_tools.cr_id(service, instance), "config"
    )

    assert config == {
        "flink-version": "1.13.5",
        "flink-revision": "0ff28a7 @ 2021-12-14T23:26:04+01:00",
    }


@mock.patch("requests.Response", autospec=True)
@mock.patch("paasta_tools.flink_tools.requests.get", autospec=True)
@mock.patch("paasta_tools.flink_tools.get_cr", autospec=True)
def test_curl_flink_endpoint_list_jobs(
    mock_get_cr,
    mock_requests_get,
    mock_response,
):
    mock_get_cr.return_value = {
        "metadata": {
            "labels": {"paasta.yelp.com/cluster": "mocked"},
            "annotations": {
                "flink.yelp.com/dashboard_url": "http://flink.k8s.test_cluster.paasta:31080/kurupt-7f5cfd8ffc"
            },
        }
    }
    mock_requests_get.return_value = mock_response
    mock_response.json.return_value = {
        "jobs": [{"id": "4210f0646f5c9ce1db0b3e5ae4372b82", "status": "RUNNING"}]
    }

    service = "kurupt"
    instance = "main"
    jobs = flink_tools.curl_flink_endpoint(flink_tools.cr_id(service, instance), "jobs")

    assert jobs == {
        "jobs": [{"id": "4210f0646f5c9ce1db0b3e5ae4372b82", "status": "RUNNING"}]
    }


@mock.patch("requests.Response", autospec=True)
@mock.patch("paasta_tools.flink_tools.requests.get", autospec=True)
@mock.patch("paasta_tools.flink_tools.get_cr", autospec=True)
def test_curl_flink_endpoint_get_job_details(
    mock_get_cr,
    mock_requests_get,
    mock_response,
):
    mock_get_cr.return_value = {
        "metadata": {
            "labels": {"paasta.yelp.com/cluster": "mocked"},
            "annotations": {
                "flink.yelp.com/dashboard_url": "http://flink.k8s.test_cluster.paasta:31080/kurupt-7f5cfd8ffc"
            },
        }
    }
    mock_requests_get.return_value = mock_response
    mock_response.json.return_value = {
        "jid": "4210f0646f5c9ce1db0b3e5ae4372b82",
        "name": "beam_happyhour.main.test_job",
        "isStoppable": False,
        "state": "RUNNING",
        "start-time": 1655053223341,
        "end-time": -1,
        "duration": 855872054,
        "maxParallelism": -1,
        "now": 1655909095395,
        "timestamps": {
            "CREATED": 1655053223735,
            "SUSPENDED": 0,
            "FAILING": 0,
            "FINISHED": 0,
            "FAILED": 0,
            "RUNNING": 1655842396454,
            "CANCELLING": 0,
            "RESTARTING": 1655842393301,
            "RECONCILING": 0,
            "CANCELED": 0,
            "INITIALIZING": 1655053223341,
        },
    }

    service = "kurupt"
    instance = "main"
    job = flink_tools.curl_flink_endpoint(
        flink_tools.cr_id(service, instance), "jobs/4210f0646f5c9ce1db0b3e5ae4372b82"
    )

    assert job == {
        "jid": "4210f0646f5c9ce1db0b3e5ae4372b82",
        "state": "RUNNING",
        "name": "beam_happyhour.main.test_job",
        "start-time": 1655053223341,
    }


def test_get_flink_jobmanager_overview():
    with mock.patch(
        "paasta_tools.flink_tools._dashboard_get",
        autospec=True,
        return_value='{"taskmanagers":10,"slots-total":10,"flink-version":"1.6.4","flink-commit":"6241481"}',
    ) as mock_dashboard_get:
        cluster = "mycluster"
        cr_name = "kurupt--fm-7c7b459d59"
        overview = flink_tools.get_flink_jobmanager_overview(cr_name, cluster, False)
        mock_dashboard_get.assert_called_once_with(
            cr_name=cr_name, cluster=cluster, path="overview", is_eks=False
        )
        assert overview == {
            "taskmanagers": 10,
            "slots-total": 10,
            "flink-version": "1.6.4",
            "flink-commit": "6241481",
        }



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock

from paasta_tools import adhoc_tools
from paasta_tools.utils import DeploymentsJsonV2
from paasta_tools.utils import NoConfigurationForServiceError


def test_get_default_interactive_config():
    with mock.patch(
        "paasta_tools.adhoc_tools.load_adhoc_job_config", autospec=True
    ) as mock_load_adhoc_job_config:
        mock_load_adhoc_job_config.return_value = adhoc_tools.AdhocJobConfig(
            service="fake_service",
            instance="interactive",
            cluster="fake_cluster",
            config_dict={},
            branch_dict={"deploy_group": "fake_deploy_group"},
        )
        result = adhoc_tools.get_default_interactive_config(
            "fake_service", "fake_cluster", "/fake/soa/dir", load_deployments=False
        )
        assert result.get_cpus() == 4
        assert result.get_mem() == 10240
        assert result.get_disk() == 1024


def test_get_default_interactive_config_reads_from_tty():
    with mock.patch(
        "paasta_tools.adhoc_tools.prompt_pick_one", autospec=True
    ) as mock_prompt_pick_one, mock.patch(
        "paasta_tools.adhoc_tools.load_adhoc_job_config", autospec=True
    ) as mock_load_adhoc_job_config, mock.patch(
        "paasta_tools.adhoc_tools.load_v2_deployments_json", autospec=True
    ) as mock_load_deployments_json:
        mock_prompt_pick_one.return_value = "fake_deploygroup"
        mock_load_adhoc_job_config.side_effect = NoConfigurationForServiceError
        mock_load_deployments_json.return_value = DeploymentsJsonV2(
            service="fake-service",
            config_dict={
                "deployments": {
                    "fake_deploygroup": {
                        "docker_image": mock.sentinel.docker_image,
                        "git_sha": mock.sentinel.git_sha,
                    }
                },
                "controls": {},
            },
        )
        result = adhoc_tools.get_default_interactive_config(
            "fake_service", "fake_cluster", "/fake/soa/dir", load_deployments=True
        )
        assert result.get_deploy_group() == "fake_deploygroup"
        assert result.get_docker_image() == mock.sentinel.docker_image



import asyncio
import functools
import weakref
from collections import defaultdict

import mock
import pytest

from paasta_tools.async_utils import async_ttl_cache


@pytest.mark.asyncio
async def test_async_ttl_cache_hit():
    return_values = iter(range(10))

    @async_ttl_cache(ttl=None)
    async def range_coroutine():
        return next(return_values)

    assert await range_coroutine() == await range_coroutine()


@pytest.mark.asyncio
async def test_async_ttl_cache_miss():
    return_values = iter(range(10))

    @async_ttl_cache(ttl=0)
    async def range_coroutine():
        return next(return_values)

    assert await range_coroutine() != await range_coroutine()


@pytest.mark.asyncio
async def test_async_ttl_cache_doesnt_cache_failures():
    flaky_error_raiser = mock.Mock(side_effect=[Exception, None])

    @async_ttl_cache(ttl=None)
    async def flaky_coroutine():
        return flaky_error_raiser()

    with pytest.raises(Exception):
        await flaky_coroutine()

    # if we were caching failures, this would fail
    assert await flaky_coroutine() is None


class DataHolder:
    def __init__(self, value):
        self.value = value


@pytest.mark.asyncio
async def test_async_ttl_cache_returns_in_flight_future():
    return_values = iter(range(10))
    event = asyncio.Event()
    condition = asyncio.Condition()
    num_waiting_coroutines = DataHolder(value=0)

    # Wait until we have enough coroutines waiting to return a result.  This
    # ensures that dependent coroutines have a chance to get a future out of
    # the cache
    @async_ttl_cache(ttl=0)
    async def range_coroutine():
        await event.wait()
        return next(return_values)

    # Wait until we have enough coroutines waiting on range_coroutine, then
    # wake range_coroutine
    async def event_setter():
        async with condition:
            while num_waiting_coroutines.value != 2:
                await condition.wait()
            event.set()

    # Keep track of how many waiting range_coroutines we have to ensure both
    # have had a chance to get the in-flight future out of the cache.  This has
    # to be separate from range_coroutine since we only end up with one
    # invocation of that method due to caching.  It also has to be separate
    # from event_setter to ensure that the event is not set until both
    # coroutines are waiting.
    async def cache_waiter():
        async with condition:
            num_waiting_coroutines.value += 1
            condition.notify_all()
        return await range_coroutine()

    event_setter_future = asyncio.ensure_future(event_setter())
    future1 = asyncio.ensure_future(cache_waiter())
    future2 = asyncio.ensure_future(cache_waiter())
    await asyncio.wait([event_setter_future, future1, future2])

    assert future1.result() == future2.result() == 0


@pytest.mark.asyncio
async def test_async_ttl_cache_dont_overwrite_new_cache_entry():
    """Make sure that we don't overwrite a new cache entry that was placed
    while we were waiting to handle the result of a previously cached future
    """
    range_continue_event = asyncio.Event()
    update_cache_event = asyncio.Event()
    return_values = iter(range(10))

    # Wait until awaiter has had a chance to get the in-flight future out of
    # the cache, then signal to the cache_updater to replace the cached future
    # before returning.  Because cache_updater is signalled first, it will
    # replace the previously cached future before async_ttl_cache decides
    # whether save the result of that future in the cache
    async def range_coroutine():
        await range_continue_event.wait()
        update_cache_event.set()
        return next(return_values)

    range_coroutine_future = asyncio.ensure_future(range_coroutine())
    cache_key = functools._make_key((), {}, typed=False)
    cache = {cache_key: (range_coroutine_future, float("Inf"))}

    cached_range_coroutine = async_ttl_cache(cache=cache, ttl=0)(range_coroutine)

    new_range_coroutine_future = asyncio.ensure_future(range_coroutine())

    async def awaiter():
        range_continue_event.set()
        await cached_range_coroutine()

    async def cache_updater():
        await update_cache_event.wait()
        cache[cache_key] = (new_range_coroutine_future, float("Inf"))

    await asyncio.gather(awaiter(), cache_updater())
    assert cache[cache_key] == (new_range_coroutine_future, float("Inf"))


@pytest.mark.asyncio
async def test_async_ttl_cache_recover_if_cache_entry_removed():
    """Ensure we handle the case where we encounter an exception in the cached
    future but another coroutine awaiting the same future ran first and alraedy
    deleted the cache entry"""
    range_continue_event = asyncio.Event()
    num_awaiters_awaiting = DataHolder(value=0)

    class TestException(Exception):
        pass

    async def range_coroutine():
        await range_continue_event.wait()
        raise TestException

    range_coroutine_future = asyncio.ensure_future(range_coroutine())
    cache_key = functools._make_key((), {}, typed=False)
    cache = {cache_key: (range_coroutine_future, float("Inf"))}

    cached_range_coroutine = async_ttl_cache(cache=cache, ttl=0)(range_coroutine)

    async def awaiter():
        num_awaiters_awaiting.value += 1
        if num_awaiters_awaiting.value == 2:
            range_continue_event.set()

        try:
            await cached_range_coroutine()
        except TestException:
            pass

    # should not raise a KeyError!
    await asyncio.gather(awaiter(), awaiter())


@pytest.mark.asyncio
async def test_async_ttl_cache_for_class_members_doesnt_leak_mem():
    """Ensure that we aren't leaking memory"""

    x = 42
    instance_caches = defaultdict(dict)

    class TestClass:
        @async_ttl_cache(ttl=None, cleanup_self=True, cache=instance_caches)
        async def f(self):
            return x

    o1 = TestClass()
    w1 = weakref.ref(o1)
    assert w1() is not None
    assert await o1.f() == x
    assert len(instance_caches) == 1
    assert list(instance_caches.keys())[0]() == o1
    del o1
    assert len(instance_caches) == 0
    assert w1() is None

    o2, o3, o4 = TestClass(), TestClass(), TestClass()
    assert await o2.f() == x
    assert await o3.f() == x
    assert await o4.f() == x
    assert len(instance_caches) == 3
    del o2, o4
    assert len(instance_caches) == 1
    del o3
    assert len(instance_caches) == 0



import mock

from paasta_tools.kubernetes_tools import registration_label
from paasta_tools.setup_istio_mesh import cleanup_paasta_namespace_services
from paasta_tools.setup_istio_mesh import sanitise_kubernetes_service_name
from paasta_tools.setup_istio_mesh import setup_istio_mesh
from paasta_tools.setup_istio_mesh import setup_paasta_namespace_services
from paasta_tools.setup_istio_mesh import setup_paasta_routing
from paasta_tools.setup_istio_mesh import UNIFIED_K8S_SVC_NAME
from paasta_tools.setup_istio_mesh import UNIFIED_SVC_PORT


MOCK_PORT_NUMBER = 20508


def test_setup_kube_service():
    mock_client = mock.Mock()
    service_name = "compute-infra-test-service.main"
    mock_paasta_namespaces = {service_name: {"port": MOCK_PORT_NUMBER}}
    sanitized_service_name = sanitise_kubernetes_service_name(service_name)

    fn, *rest = setup_paasta_namespace_services(
        kube_client=mock_client,
        paasta_namespaces=mock_paasta_namespaces,
        existing_namespace_services={},
        existing_virtual_services={},
    )

    k8s_svc = fn.args[1]

    assert len(rest) == 1
    assert fn.func is mock_client.core.create_namespaced_service
    assert k8s_svc.metadata.name == sanitized_service_name
    assert k8s_svc.spec.selector == {registration_label(service_name): "true"}


def test_setup_paasta_routing():
    mock_client = mock.Mock()
    mock_namespaces = dict(foo=dict(proxy_port=MOCK_PORT_NUMBER))

    fn, *rest = setup_paasta_routing(
        kube_client=mock_client,
        namespaces=mock_namespaces,
    )
    k8s_svc = fn.args[1]

    assert len(rest) == 1
    assert fn.func is mock_client.core.create_namespaced_service
    assert len(k8s_svc.spec.ports) == 2
    assert k8s_svc.spec.ports[0].port == UNIFIED_SVC_PORT
    assert k8s_svc.spec.ports[1].port == MOCK_PORT_NUMBER


def test_cleanup_paasta_namespace_services_does_not_remove_unified_svc():
    mock_client = mock.Mock()
    mock_paasta_namespaces = {"svc1", "svc2"}

    mock_existing_namespace_services = {"svc1", "svc2", "svc3", UNIFIED_K8S_SVC_NAME}
    calls = list(
        cleanup_paasta_namespace_services(
            mock_client,
            mock_paasta_namespaces,
            mock_existing_namespace_services,
            mock_existing_namespace_services,
        )
    )
    funcs = {fn.func for fn in calls}
    services = {calls[0].args[0], calls[1].args[-1]}

    assert {
        mock_client.core.delete_namespaced_service,
        mock_client.custom.delete_namespaced_custom_object,
    } == funcs
    assert {"svc3"} == services


def test_cleanup_paasta_namespace_services_does_not_remove_svc_while_running_first_time():
    mock_client = mock.Mock()
    mock_paasta_namespaces = {"svc1", "svc2"}

    mock_existing_namespace_services = {}
    calls = cleanup_paasta_namespace_services(
        mock_client,
        mock_paasta_namespaces,
        mock_existing_namespace_services,
        mock_existing_namespace_services,
    )
    assert len(list(calls)) == 0


@mock.patch("time.sleep", autospec=True)
@mock.patch("paasta_tools.setup_istio_mesh.process_kube_services", autospec=True)
def test_setup_istio_mesh_iterates(mock_process_kube_services, mock_sleep):
    mock_yielded = mock.Mock()
    mock_process_kube_services.return_value = iter([mock_yielded])
    setup_istio_mesh(mock.Mock())
    assert len(mock_yielded.mock_calls) == 1



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
PaaSTA service list (instances) etc.
"""
import json
import logging
from datetime import datetime
from datetime import timezone
from typing import Tuple

from kubernetes.client import V1ConfigMap
from pyramid.response import Response
from pyramid.view import view_config

from paasta_tools.api import settings
from paasta_tools.api.views.exception import ApiFailure
from paasta_tools.cli.utils import get_instance_config
from paasta_tools.kubernetes_tools import AUTOSCALING_OVERRIDES_CONFIGMAP_NAME
from paasta_tools.kubernetes_tools import AUTOSCALING_OVERRIDES_CONFIGMAP_NAMESPACE
from paasta_tools.kubernetes_tools import get_or_create_namespaced_configmap
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.kubernetes_tools import patch_namespaced_configmap


log = logging.getLogger(__name__)


@view_config(route_name="service.autoscaler.get", request_method="GET", renderer="json")
def get_autoscaler_count(request):
    service = request.swagger_data.get("service")
    instance = request.swagger_data.get("instance")
    cluster = settings.cluster
    soa_dir = settings.soa_dir

    instance_config = get_instance_config(service, instance, cluster, soa_dir)
    if not isinstance(instance_config, (KubernetesDeploymentConfig)):
        error_message = (
            f"Autoscaling is not supported for {service}.{instance} because instance type is not "
            f"kubernetes."
        )
        raise ApiFailure(error_message, 501)

    response_body = {
        "desired_instances": instance_config.get_instances(),
        "calculated_instances": instance_config.get_instances(with_limit=False),
    }
    return Response(json_body=response_body, status_code=200)


@view_config(
    route_name="service.autoscaler.post", request_method="POST", renderer="json"
)
def update_autoscaler_count(request):
    service = request.swagger_data.get("service")
    instance = request.swagger_data.get("instance")
    cluster = settings.cluster
    soa_dir = settings.soa_dir
    desired_instances = request.swagger_data.get("json_body")["desired_instances"]
    if not isinstance(desired_instances, int):
        error_message = 'The provided body does not have an integer value for "desired_instances": {}'.format(
            request.swagger_data.get("json_body")
        )
        raise ApiFailure(error_message, 500)

    instance_config = get_instance_config(service, instance, cluster, soa_dir, True)
    if not isinstance(instance_config, (KubernetesDeploymentConfig)):
        error_message = (
            f"Autoscaling is not supported for {service}.{instance} because instance type is not "
            f"kubernetes."
        )
        raise ApiFailure(error_message, 501)

    max_instances = instance_config.get_max_instances()
    if max_instances is None:
        error_message = f"Autoscaling is not enabled for {service}.{instance}"
        raise ApiFailure(error_message, 404)
    min_instances = instance_config.get_min_instances()

    status = "SUCCESS"
    if desired_instances > max_instances:
        desired_instances = max_instances
        status = (
            "WARNING desired_instances is greater than max_instances %d" % max_instances
        )
    elif desired_instances < min_instances:
        desired_instances = min_instances
        status = (
            "WARNING desired_instances is less than min_instances %d" % min_instances
        )
    try:
        if isinstance(instance_config, KubernetesDeploymentConfig):
            instance_config.set_autoscaled_instances(
                instance_count=desired_instances, kube_client=settings.kubernetes_client
            )
        else:
            instance_config.set_autoscaled_instances(instance_count=desired_instances)
    except Exception as err:
        raise ApiFailure(err, 500)

    response_body = {"desired_instances": desired_instances, "status": status}
    return Response(json_body=response_body, status_code=202)


def get_or_create_autoscaling_overrides_configmap() -> Tuple[V1ConfigMap, bool]:
    return get_or_create_namespaced_configmap(
        AUTOSCALING_OVERRIDES_CONFIGMAP_NAME,
        namespace=AUTOSCALING_OVERRIDES_CONFIGMAP_NAMESPACE,
        kube_client=settings.kubernetes_client,
    )


@view_config(
    route_name="service_autoscaler.autoscaling_override.post",
    request_method="POST",
    renderer="json",
)
def set_autoscaling_override(request):
    """Set a temporary autoscaling override for a service/instance.

    This endpoint creates or updates a ConfigMap entry with override information
    including expiration time. The override will be applied by the autoscaler.

    Required parameters:
    - service: The service name
    - instance: The instance name
    - min_instances: The minimum number of instances to enforce
    - expires_after: unix timestamp after which the override is no longer valid
    """
    service = request.swagger_data.get("service")
    instance = request.swagger_data.get("instance")
    cluster = settings.cluster
    soa_dir = settings.soa_dir

    instance_config = get_instance_config(
        service, instance, cluster, soa_dir, load_deployments=False
    )
    if not isinstance(instance_config, KubernetesDeploymentConfig):
        error_message = (
            f"Autoscaling is not supported for {service}.{instance} because instance type is not "
            f"kubernetes."
        )
        raise ApiFailure(error_message, 501)

    json_body = request.swagger_data.get("json_body", {})
    min_instances_override = json_body.get("min_instances")
    expire_after = json_body.get("expire_after")

    if not isinstance(min_instances_override, int) or min_instances_override < 1:
        raise ApiFailure("min_instances must be a positive integer", 400)

    if not expire_after:
        raise ApiFailure("expire_after is required", 400)

    max_instances = instance_config.get_max_instances()
    if max_instances is None:
        raise ApiFailure(f"Autoscaling is not enabled for {service}.{instance}", 400)

    if max_instances < min_instances_override:
        raise ApiFailure(
            f"min_instances ({min_instances_override}) cannot be greater than max_instances ({max_instances})",
            400,
        )

    configmap, created = get_or_create_autoscaling_overrides_configmap()
    if created:
        log.info("Created new autoscaling overrides ConfigMap")
    # i dunno why this is necessary, but a newly created configmap doesn't have a data field
    # even when we set it in the create call
    if not configmap.data:
        configmap.data = {}

    override_data = {
        "min_instances": min_instances_override,
        "created_at": datetime.now(timezone.utc).isoformat(),
        # NOTE: we may want to also allow setting a max_instances override in the future, but if we do that
        # we'd probably want to force folks to either set one or both and share the same expiration time
        "expire_after": expire_after,
    }

    service_instance = f"{service}.{instance}"
    existing_overrides = (
        json.loads(configmap.data[service_instance])
        if service_instance in configmap.data
        else {}
    )
    merged_overrides = {**existing_overrides, **override_data}
    serialized_overrides = json.dumps(merged_overrides)

    patch_namespaced_configmap(
        name=AUTOSCALING_OVERRIDES_CONFIGMAP_NAME,
        namespace=AUTOSCALING_OVERRIDES_CONFIGMAP_NAMESPACE,
        # this should only update the single entry for the $service.$instance key
        # ain't k8s grand?
        body={"data": {service_instance: serialized_overrides}},
        kube_client=settings.kubernetes_client,
    )

    response_body = {
        "service": service,
        "instance": instance,
        "cluster": cluster,
        "min_instances": min_instances_override,
        "expire_after": expire_after,
        "status": "SUCCESS",
    }
    # NOTE: this is an HTTP 202 since actually updating the HPA happens asynchronously
    # through setup_kubernetes_job
    # XXX: should we try to patch things here as well?
    return Response(json_body=response_body, status_code=202)



from collections import namedtuple

import iptc
import mock
import pytest

from paasta_tools import iptables


EMPTY_RULE = iptables.Rule(
    protocol="ip",
    src="0.0.0.0/0.0.0.0",
    dst="0.0.0.0/0.0.0.0",
    target=None,
    matches=(),
    target_parameters=(),
)


@pytest.fixture
def mock_Table():
    with mock.patch.object(iptc, "Table", autospec=True) as m:
        m.return_value.autocommit = True
        yield m


@pytest.fixture
def mock_Chain():
    with mock.patch.object(iptc, "Chain", autospec=True) as m:
        yield m


def test_rule_from_iptc_simple():
    rule = iptc.Rule()
    rule.create_target("DROP")
    rule.src = "169.229.226.0/255.255.255.0"

    assert iptables.Rule.from_iptc(rule) == EMPTY_RULE._replace(
        src="169.229.226.0/255.255.255.0", target="DROP"
    )


def test_rule_from_iptc_mac_match():
    rule = iptc.Rule()
    rule.create_target("DROP")
    rule.create_match("mac")
    rule.matches[0].mac_source = "20:c9:d0:2b:6f:f3"

    iptables_rule = iptables.Rule.from_iptc(rule)
    assert iptables_rule == EMPTY_RULE._replace(
        target="DROP", matches=(("mac", (("mac-source", ("20:c9:d0:2b:6f:f3",)),)),)
    ) or iptables_rule == EMPTY_RULE._replace(
        target="DROP", matches=(("mac", (("mac-source", ("20:C9:D0:2B:6F:F3",)),)),)
    )


def test_rule_from_iptc_target_parameters():
    rule = iptc.Rule()
    target = rule.create_target("LOG")
    target.set_parameter("log-prefix", "my-prefix ")

    assert iptables.Rule.from_iptc(rule) == EMPTY_RULE._replace(
        target="LOG", target_parameters=(("log-prefix", ("my-prefix ",)),)
    )


def test_rule_tcp_to_iptc():
    rule = EMPTY_RULE._replace(
        protocol="tcp", target="ACCEPT", matches=(("tcp", (("dport", ("443",)),)),)
    ).to_iptc()
    assert rule.protocol == "tcp"
    assert rule.target.name == "ACCEPT"
    assert len(rule.matches) == 1
    assert rule.matches[0].name == "tcp"
    assert rule.matches[0].parameters["dport"] == "443"


def test_mac_src_to_iptc():
    rule = EMPTY_RULE._replace(
        target="ACCEPT", matches=(("mac", (("mac-source", ("20:C9:D0:2B:6F:F3",)),)),)
    ).to_iptc()
    assert rule.protocol == "ip"
    assert rule.target.name == "ACCEPT"
    assert len(rule.matches) == 1
    assert rule.matches[0].name == "mac"
    assert rule.matches[0].parameters["mac_source"].upper() == "20:C9:D0:2B:6F:F3"


def test_iptables_txn_normal():
    table = mock.Mock(autocommit=True)
    with iptables.iptables_txn(table):
        assert table.autocommit is False
        assert table.commit.called is False
        assert table.refresh.called is False
    assert table.commit.called is True
    assert table.refresh.called is True
    assert table.autocommit is True


def test_iptables_txn_with_exception():
    table = mock.Mock(autocommit=True)
    with pytest.raises(ValueError):
        with iptables.iptables_txn(table):
            raise ValueError("just testing lol")
    assert table.commit.called is False
    assert table.refresh.called is True
    assert table.autocommit is True


def test_all_chains(mock_Table):
    chain1 = mock.Mock()
    chain1.name = "INPUT"
    chain2 = mock.Mock()
    chain2.name = "OUTPUT"
    mock_Table.return_value = mock.Mock(chains=[chain1, chain2])
    assert iptables.all_chains() == {"INPUT", "OUTPUT"}


def test_ensure_chain():
    with mock.patch.object(
        iptables,
        "list_chain",
        autospec=True,
        return_value={
            EMPTY_RULE._replace(target="DROP"),
            EMPTY_RULE._replace(target="ACCEPT", src="1.0.0.0/255.255.255.0"),
        },
    ), mock.patch.object(
        iptables, "insert_rule", autospec=True
    ) as mock_insert_rule, mock.patch.object(
        iptables, "delete_rules", autospec=True
    ) as mock_delete_rules:
        iptables.ensure_chain(
            "PAASTA.service",
            (
                EMPTY_RULE._replace(target="DROP"),
                EMPTY_RULE._replace(target="ACCEPT", src="2.0.0.0/255.255.255.0"),
            ),
        )

    # It should add the missing rule
    assert mock_insert_rule.mock_calls == [
        mock.call(
            "PAASTA.service",
            EMPTY_RULE._replace(target="ACCEPT", src="2.0.0.0/255.255.255.0"),
        )
    ]

    # It should delete the extra rule
    assert mock_delete_rules.mock_calls == [
        mock.call(
            "PAASTA.service",
            {EMPTY_RULE._replace(target="ACCEPT", src="1.0.0.0/255.255.255.0")},
        )
    ]


def test_ensure_chain_creates_chain_if_doesnt_exist():
    with mock.patch.object(
        iptables, "list_chain", side_effect=iptables.ChainDoesNotExist("PAASTA.service")
    ), mock.patch.object(iptables, "create_chain", autospec=True) as mock_create_chain:
        iptables.ensure_chain("PAASTA.service", ())

    assert mock_create_chain.mock_calls == [mock.call("PAASTA.service")]


def test_ensure_rule_does_not_exist():
    with mock.patch.object(
        iptables,
        "list_chain",
        return_value=(
            EMPTY_RULE._replace(target="ACCEPT"),
            EMPTY_RULE._replace(src="10.0.0.0/255.255.255.0"),
        ),
    ), mock.patch.object(iptables, "insert_rule", autospec=True) as mock_insert_rule:
        iptables.ensure_rule("PAASTA.service", EMPTY_RULE._replace(target="DROP"))

    assert mock_insert_rule.mock_calls == [
        mock.call("PAASTA.service", EMPTY_RULE._replace(target="DROP"))
    ]


def test_ensure_rule_already_exists():
    with mock.patch.object(
        iptables,
        "list_chain",
        return_value=(
            EMPTY_RULE._replace(target="DROP"),
            EMPTY_RULE._replace(src="10.0.0.0/255.255.255.0"),
        ),
    ), mock.patch.object(iptables, "insert_rule", autospec=True) as mock_insert_rule:
        iptables.ensure_rule("PAASTA.service", EMPTY_RULE._replace(target="DROP"))

    assert mock_insert_rule.called is False


def test_insert_rule(mock_Table, mock_Chain):
    iptables.insert_rule("PAASTA.service", EMPTY_RULE._replace(target="DROP"))

    (call,) = mock_Chain("filter", "PAASTA.service").insert_rule.call_args_list
    args, kwargs = call
    (rule,) = args
    assert iptables.Rule.from_iptc(rule) == EMPTY_RULE._replace(target="DROP")


def test_delete_rules(mock_Table, mock_Chain):
    mock_Chain.return_value.rules = (
        EMPTY_RULE._replace(target="DROP").to_iptc(),
        EMPTY_RULE._replace(target="ACCEPT").to_iptc(),
        EMPTY_RULE._replace(
            target="REJECT",
            target_parameters=(("reject-with", ("icmp-port-unreachable",)),),
        ).to_iptc(),
    )
    iptables.delete_rules(
        "PAASTA.service",
        (
            EMPTY_RULE._replace(target="ACCEPT"),
            EMPTY_RULE._replace(
                target="REJECT",
                target_parameters=(("reject-with", ("icmp-port-unreachable",)),),
            ),
        ),
    )
    assert mock_Chain("filter", "PAASTA.service").delete_rule.mock_calls == [
        mock.call(mock_Chain.return_value.rules[1]),
        mock.call(mock_Chain.return_value.rules[2]),
    ]


def test_create_chain(mock_Table):
    iptables.create_chain("PAASTA.service")
    mock_Table("filter").create_chain.assert_called_once_with("PAASTA.service")


def test_delete_chain(mock_Table, mock_Chain):
    iptables.delete_chain("PAASTA.service")
    chain = mock_Chain("filter", "PAASTA.service")
    assert chain.flush.called is True
    assert chain.delete.called is True


def test_list_chain_simple(mock_Table, mock_Chain):
    chain = mock_Chain("PAASTA.internet", mock_Table.return_value)
    rule = iptc.Rule()
    rule.create_target("DROP")
    chain.rules = [rule]
    mock_Table.return_value.chains = [chain]
    assert iptables.list_chain("PAASTA.internet") == (
        EMPTY_RULE._replace(target="DROP"),
    )


def test_list_chain_does_not_exist(mock_Table, mock_Chain):
    mock_Table.return_value.chains = []
    with pytest.raises(iptables.ChainDoesNotExist):
        iptables.list_chain("PAASTA.internet")


class TestReorderChain:
    class FakeRule(namedtuple("FakeRule", ("target", "id"))):
        def to_iptc(self):
            return self

    @pytest.fixture(autouse=True)
    def chain_mock(self):
        with mock.patch.object(
            iptables, "iptables_txn", autospec=True
        ), mock.patch.object(iptables.iptc, "Table", autospec=True), mock.patch.object(
            iptables.iptc, "Chain", autospec=True
        ) as chain_mock, mock.patch.object(
            iptables, "list_chain", autospec=True
        ) as list_chain_mock:
            self.chain_mock = chain_mock
            self.list_chain_mock = list_chain_mock
            yield

    def test_reorder_chain_flip(self):
        self.list_chain_mock.return_value = [
            self.FakeRule("REJECT", "a"),
            self.FakeRule("LOG", "b"),
            self.FakeRule("ACCEPT", "c"),
            self.FakeRule("ACCEPT", "d"),
        ]
        iptables.reorder_chain("")
        assert self.chain_mock.return_value.replace_rule.mock_calls == [
            mock.call(self.FakeRule("ACCEPT", "c"), 0),
            mock.call(self.FakeRule("ACCEPT", "d"), 1),
            mock.call(self.FakeRule("LOG", "b"), 2),
            mock.call(self.FakeRule("REJECT", "a"), 3),
        ]

    def test_reorder_chain_log_first(self):
        self.list_chain_mock.return_value = [
            self.FakeRule("LOG", "b"),
            self.FakeRule("ACCEPT", "c"),
            self.FakeRule("ACCEPT", "d"),
            self.FakeRule("REJECT", "a"),
        ]
        iptables.reorder_chain("")
        assert self.chain_mock.return_value.replace_rule.mock_calls == [
            mock.call(self.FakeRule("ACCEPT", "c"), 0),
            mock.call(self.FakeRule("ACCEPT", "d"), 1),
            mock.call(self.FakeRule("LOG", "b"), 2),
        ]

    def test_reorder_chain_empty(self):
        self.list_chain_mock.return_value = []
        iptables.reorder_chain("")
        assert self.chain_mock.return_value.replace_rule.mock_calls == []

    def test_reorder_chain_already_in_order(self):
        self.chain_mock.return_value.rules = [
            self.FakeRule("ACCEPT", "c"),
            self.FakeRule("ACCEPT", "d"),
            self.FakeRule("LOG", "b"),
            self.FakeRule("REJECT", "a"),
        ]
        iptables.reorder_chain("")
        assert self.chain_mock.return_value.replace_rule.mock_calls == []

    def test_reorder_chain_log_at_bottom(self):
        self.list_chain_mock.return_value = [
            self.FakeRule("ACCEPT", "c"),
            self.FakeRule("ACCEPT", "d"),
            self.FakeRule("REJECT", "a"),
            self.FakeRule("LOG", "b"),
        ]
        iptables.reorder_chain("")
        assert self.chain_mock.return_value.replace_rule.mock_calls == [
            mock.call(self.FakeRule("LOG", "b"), 2),
            mock.call(self.FakeRule("REJECT", "a"), 3),
        ]

    def test_reorder_chain_reject_in_middle(self):
        self.list_chain_mock.return_value = [
            self.FakeRule("ACCEPT", "c"),
            self.FakeRule("REJECT", "a"),
            self.FakeRule("ACCEPT", "d"),
        ]
        iptables.reorder_chain("")
        assert self.chain_mock.return_value.replace_rule.mock_calls == [
            mock.call(self.FakeRule("ACCEPT", "d"), 1),
            mock.call(self.FakeRule("REJECT", "a"), 2),
        ]

    def test_reorder_chain_other_target_names(self):
        self.list_chain_mock.return_value = [
            self.FakeRule("HELLOWORLD", "c"),
            self.FakeRule("REJECT", "a"),
            self.FakeRule("FOOBAR", "d"),
        ]
        iptables.reorder_chain("")
        assert self.chain_mock.return_value.replace_rule.mock_calls == [
            mock.call(self.FakeRule("FOOBAR", "d"), 1),
            mock.call(self.FakeRule("REJECT", "a"), 2),
        ]
